#!/usr/bin/perl
# ipdb/cgi-bin/admin.cgi
# Hack interface to make specific changes to IPDB that (for one reason
# or another) can't be made through the main interface.
#
###
# SVN revision info
# $Date: 2010-09-20 21:16:54 +0000 (Mon, 20 Sep 2010) $
# SVN revision $Rev: 483 $
# Last update by $Author: kdeugau $
###
# Copyright (C) 2004-2010 - Kris Deugau

use strict;
use warnings;
use CGI::Carp qw(fatalsToBrowser);
use CGI::Simple;
use HTML::Template;
use DBI;
use CommonWeb qw(:ALL);
use CustIDCK;
#use POSIX qw(ceil);
use NetAddr::IP;

use Sys::Syslog;

# don't remove!  required for GNU/FHS-ish install from tarball
##uselib##

use MyIPDB;

openlog "IPDB-admin","pid","$IPDB::syslog_facility";

# Collect the username from HTTP auth.  If undefined, we're in a test environment.
my $authuser;
if (!defined($ENV{'REMOTE_USER'})) {
  $authuser = '__temptest';
} else {
  $authuser = $ENV{'REMOTE_USER'};
}

syslog "debug", "$authuser active";

# Why not a global DB handle?  (And a global statement handle, as well...)
# Use the connectDB function, otherwise we end up confusing ourselves
my $ip_dbh;
my $sth;
my $errstr;
($ip_dbh,$errstr) = connectDB_My;
if (!$ip_dbh) {
  printAndExit("Database error: $errstr\n");
}
initIPDBGlobals($ip_dbh);

# anyone got a better name?  :P
my $thingroot = $ENV{SCRIPT_FILENAME};
$thingroot =~ s|cgi-bin/admin.cgi||;

# Set up some globals
$ENV{HTML_TEMPLATE_ROOT} = $thingroot."templates";

if ($IPDBacl{$authuser} !~ /A/) {
  my $page = HTML::Template->new(filename => "admin/aclerr.tmpl");
##fixme:  need params for IPDB admin email and name
  $page->param(ipdbadmin_email => 'ipdbadmin@example.com');
  $page->param(ipdbadmin_name => 'the IPDB administrator');
  print "Content-Type: text/html\n\n".$page->output;
  exit;
}

# Set up the CGI object...
my $q = new CGI::Simple;
# ... and get query-string params as well as POST params if necessary
$q->parse_query_string;

# Convenience;  saves changing all references to %webvar
##fixme:  tweak for handling <select multiple='y' size=3> (list with multiple selection)
my %webvar = $q->Vars;

my $header = HTML::Template->new(filename => "admin/header.tmpl");

if(!defined($webvar{action})) {
  $webvar{action} = "main";   #shuts up the warnings.
}

#print "Content-type: text/html\n\n".$header->output;
my $page = HTML::Template->new(filename => "admin/$webvar{action}.tmpl");

if ($webvar{action} eq 'main') {
  $header->param(mainpage => 1);

  $sth = $ip_dbh->prepare("select type,listname from alloctypes where listorder < 900 order by listorder");
  $sth->execute;

  my @typelist;
  my $count = 0;
  while (my ($type,$listname) = $sth->fetchrow_array) {
    my %row = (
	selected => $count++,
	type => $type,
	dispname => $listname
	);
    push @typelist, \%row;
  }
  $page->param(typelist => \@typelist);

  my @masterlist;
  $sth = $ip_dbh->prepare("select cidr,mtime from masterblocks order by cidr");
  $sth->execute;
  while (my ($cidr,$mtime) = $sth->fetchrow_array) {
    my %row = (
	master => $cidr,
	masterdate => $mtime
	);
    push @masterlist, \%row;
  }
  $page->param(masterlist => \@masterlist);

#} else {
#  print '<a href="/ip/cgi-bin/admin.cgi">Back</a> to main<hr>';
}


## Possible actions.
elsif ($webvar{action} eq 'alloc') {
  # OK, we know what we're allocating.

  if ($webvar{cidr} !~ /^\s*(\d{1,3}\.){3}\d{1,3}(\/\d{2})?\s*$/) {
    printAndExit("Can't allocate something that's not a netblock/ip");
  }

  $sth = $ip_dbh->prepare("select def_custid from alloctypes where type='$webvar{alloctype}'");
  $sth->execute;
  my @data = $sth->fetchrow_array;
  my $custid = $data[0];
  if ($custid eq '') {
    if ($webvar{custid} !~ /^(?:\d{10}|\d{7}|STAFF)(?:-\d\d?)?$/) {
      # Force uppercase for now...
      $webvar{custid} =~ tr/a-z/A-Z/;
      # Crosscheck with billing.
      my $status = CustIDCK->custid_exist($webvar{custid});
      if ($CustIDCK::Error) {
	printError("Error verifying customer ID: ".$CustIDCK::ErrMsg);
	return;
      }
      if (!$status) {
	printError("Customer ID not valid.  Make sure the Customer ID ".
	  "is correct.<br>\nUse STAFF for staff static IPs, and $IPDB::defcustid for any other ".
	  "non-customer assignments.");
	return;
      }
    }
    # Type that doesn't have a default custid
    $custid = $webvar{custid};
  }

  my $cidr = new NetAddr::IP $webvar{cidr};
  my @data;
  if ($webvar{alloctype} eq 'rm') {
    $sth = $ip_dbh->prepare("select cidr from freeblocks where cidr >>='$cidr' and routed='n'");
    $sth->execute;
    @data = $sth->fetchrow_array;
# User deserves errors if user can't be bothered to find the free block first.
    printAndExit("Can't allocate from outside a free block!!\n")
        if !$data[0];
  } elsif ($webvar{alloctype} =~ /^(.)i$/) {
    $sth = $ip_dbh->prepare("select cidr from allocations where cidr >>='$cidr' and (type like '_d' or type like '_p')");
    $sth->execute;
    @data = $sth->fetchrow_array;
# User deserves errors if user can't be bothered to find the pool and a free IP first.
    printAndExit("Can't allocate static IP from outside a pool!!\n")
	if !$data[0];
  } else {
    $sth = $ip_dbh->prepare("select cidr from freeblocks where cidr >>='$cidr' and not (routed='n')");
    $sth->execute;
    @data = $sth->fetchrow_array;
# User deserves errors if user can't be bothered to find the free block first.
    printAndExit("Can't allocate from outside a routed block!!\n")
        if !$data[0];
  }

  my $alloc_from = new NetAddr::IP $data[0];
  $sth->finish;

  my $cities = '';
  foreach my $city (@citylist) {
    $cities .= "<option>$city</option>\n";
  }

  print qq(<table class=regular>
<form method=POST action=admin.cgi>
<tr class=color1>
<td>Allocating:</td>
<td>$cidr<input type=hidden name=cidr value="$cidr"></td>
</tr><tr class=color2>
<td>Type:</td><td>$disp_alloctypes{$webvar{alloctype}}
<input type=hidden name=alloctype value="$webvar{alloctype}"></td>
</tr><tr class=color1>
<td>Allocated from:</td>
<td>$alloc_from<input type=hidden name=alloc_from value="$alloc_from"></td>
</tr><tr class="color2">
<td>Customer ID:</td><td>$custid<input type=hidden name=custid value="$custid"></td>
</tr><tr class=color1>
<td>Customer location:</td><td>
<select name="city"><option selected>-</option>
$cities
</select>
&nbsp;<a href="javascript:popNotes('/ip/cgi-bin/newcity.cgi')">Add new location</a>
</td>
</tr>
<tr class="color2">
<td>Circuit ID:</td><td><input name=circid size=40></td>
</tr><tr class="color1">
<td>Description/Name:</td><td><input name="desc" size=40></td>
</tr><tr class="color2">
<td>Notes:</td><td><textarea name="notes" rows="3" cols="40"></textarea></td>
</tr><tr class="warning">
<td colspan=2><center>WARNING:  This will IMMEDIATELY assign this block!!</center></td>
</tr><tr class="color2">
<td class="center" colspan="2"><input type="submit" value="  Assign  "></td>
<input type="hidden" name="action" value="confirm">
</form>
</tr>
</table>
);


} elsif ($webvar{action} eq 'confirm') {

  print "Assigning $webvar{cidr} to $webvar{custid} (\"$webvar{desc}\") as ".
	"$disp_alloctypes{$webvar{alloctype}}...<br>\n";
  # Only need to check city here.
  if ($webvar{city} eq '-') {
    printError("Invalid customer location!  Go back and select customer's location.");
  } else {
    if ($webvar{alloctype} =~ /^.i$/) {
      $sth = $ip_dbh->prepare("update poolips set available='n', custid='$webvar{custid}', ".
	"city='$webvar{city}', description='$webvar{desc}', notes='$webvar{notes}' ".
	"where ip='$webvar{cidr}'");
      $sth->execute;
      if ($sth->err) {
	print "Allocation failed!  DBI said:\n".$sth->errstr."\n";
        syslog "err", "($authuser) Allocation of '$webvar{cidr}' to '$webvar{custid}' as ".
		"'$webvar{alloctype}' failed: '".$sth->errstr."'";
      } else {
	print "Allocation OK!\n";
	syslog "notice", "$authuser allocated '$webvar{cidr}' to '$webvar{custid}' as ".
		"'$webvar{alloctype}'";
	mailNotify($ip_dbh, "a$webvar{alloctype}", "ADDED: $disp_alloctypes{$webvar{alloctype}} allocation",
	  "$disp_alloctypes{$webvar{alloctype}} $webvar{cidr} allocated to customer $webvar{custid}\n".
	  "Description: $webvar{desc}\n\nAllocated by: $authuser\n");
      }
    } else {
      my ($retcode,$msg) = allocateBlock($ip_dbh, $webvar{cidr}, $webvar{alloc_from},
	$webvar{custid}, $webvar{alloctype}, $webvar{city}, $webvar{desc}, $webvar{notes},
	$webvar{circid});
      if ($retcode eq 'OK') {
	print "Allocation OK!\n";
	syslog "notice", "$authuser allocated '$webvar{cidr}' to '$webvar{custid}' as ".
		"'$webvar{alloctype}'";
      } else {
	print "Allocation failed!  IPDB::allocateBlock said:\n$msg\n";
        syslog "err", "($authuser) Allocation of '$webvar{cidr}' to '$webvar{custid}' as ".
		"'$webvar{alloctype}' failed: '$msg'";
      }
    } # static IP vs netblock

  } # done city check

} elsif ($webvar{action} eq 'alloctweak') {
  fix_allocfrom();
  showAllocs($webvar{allocfrom});
} elsif ($webvar{action} eq 'update') {
  update();
} elsif ($webvar{action} eq 'assign') {
  # Display a list of possible blocks within the requested block.
  open (HTML, "../admin_alloc.html")
	or croak "Could not open admin_alloc.html :$!";
  my $html = join('', <HTML>);
  $html =~ s/\$\$MASK\$\$/$webvar{masklen}/g;
  $html =~ s/\$\$ALLOCFROM\$\$/$webvar{allocfrom}/g;

  my $from = new NetAddr::IP $webvar{allocfrom};
  my @blocklist = $from->split($webvar{masklen});
  my $availblocks;
  foreach (@blocklist) {
    $availblocks .= qq(<tr><td colspan=2 align=center><input type=radio name=block value="$_">$_</td></tr>\n);
  }
  $html =~ s/\$\$BLOCKLIST\$\$/$availblocks/g;

  print $html;
} elsif ($webvar{action} eq 'touch') {
  print "Touching master $webvar{whichmaster}\n";
  $sth = $ip_dbh->prepare("update masterblocks set mtime=now() where cidr='$webvar{whichmaster}'");
  $sth->execute;
  if ($sth->err) {
    print "<p>Error updating modified timestamp on master $webvar{whichmaster}: ".$sth->errstr."\n";
  }

} elsif ($webvar{action} eq 'listcust') {

  $sth = $ip_dbh->prepare("select custid,name,tech_handle from customers order by custid");
  $sth->execute;
  my @custlist;
  while (my @data = $sth->fetchrow_array) {
    my %row = (
	custid => $data[0],
	custname => $data[1],
	tech => $data[2]
	);
    push @custlist, \%row;
  }
  $page->param(custlist => \@custlist);

} elsif ($webvar{action} eq 'edcust') {

  if ($webvar{newcust}) {
    $sth = $ip_dbh->prepare("INSERT INTO customers (custid) VALUES (?)");
    $sth->execute($webvar{custid});
  }
  $sth = $ip_dbh->prepare("select custid,name,street,city,province,".
	"country,pocode,phone,tech_handle,abuse_handle,admin_handle,special ".
	"from customers where custid='$webvar{custid}'");
  $sth->execute;
  my ($custid, $name, $street, $city, $prov, $country, $pocode, $phone, $tech, $abuse, $admin, $special) =
	$sth->fetchrow_array;

  $page->param(
	custid => $custid,
	name => $name,
	street => $street,
	city => $city,
	prov => $prov,
	country => $country,
	pocode => $pocode,
	phone => $phone,
	tech => $tech,
	abuse => $abuse,
	admin => $admin,
	special => $special
	);

} elsif ($webvar{action} eq 'updcust') {

  $sth = $ip_dbh->prepare("UPDATE customers SET".
	" name=?, street=?, city=?, province=?, country=?, pocode=?,".
	" phone=?, tech_handle=?, abuse_handle=?, admin_handle=?, special=?".
	" WHERE custid=?");
  $sth->execute($webvar{name}, $webvar{street}, $webvar{city}, $webvar{province}, 
	$webvar{country}, $webvar{pocode}, $webvar{phone}, $webvar{tech_handle}, 
	$webvar{abuse_handle}, $webvar{admin_handle}, $webvar{special}, $webvar{custid});
  $page->param(
	custid => $webvar{custid},
	name => $webvar{name},
	street => $webvar{street},
	city => $webvar{city},
	prov => $webvar{province},
	country => $webvar{country},
	pocode => $webvar{pocode},
	phone => $webvar{phone},
	tech => $webvar{tech_handle},
	abuse => $webvar{abuse_handle},
	admin => $webvar{admin_handle},
	special => $webvar{special}
	);

} elsif ($webvar{action} eq 'showpools') {
  print "IP Pools currently allocated:\n".
	"<table border=1>\n<tr><td>Pool</td><td># of free IPs</td></tr>\n";
  $sth = $ip_dbh->prepare("select cidr from allocations where type like '%p' or type like '%d' order by cidr");
  $sth->execute;
  my %poolfree;
  while (my @data = $sth->fetchrow_array) {
    $poolfree{$data[0]} = 0;
  }
  $sth = $ip_dbh->prepare("select pool,ip from poolips where available='y' order by ip");
  $sth->execute;
  while (my @data = $sth->fetchrow_array) {
    $poolfree{$data[0]}++;
  }
  foreach my $key (keys %poolfree) {
    print qq(<tr><td><a href="admin.cgi?action=tweakpool&pool=$key">$key</a></td>).
	"<td>$poolfree{$key}</td></tr>\n";
  }
  print "</table>\n";
} elsif ($webvar{action} eq 'tweakpool') {
  showPool($webvar{pool});
} elsif ($webvar{action} eq 'updatepool') {

  $sth = $ip_dbh->prepare("update poolips set custid='$webvar{custid}', ".
	"city='$webvar{city}', type='$webvar{type}', available='".
	(($webvar{available} eq 'y') ? 'y' : 'n').
	"', notes='$webvar{notes}', description='$webvar{desc}' ".
	"where ip='$webvar{ip}'");
  $sth->execute;
  if ($sth->err) {
    print "Error updating pool IP $webvar{ip}: $@<hr>\n";
    syslog "err", "$authuser could not update pool IP $webvar{ip}: $@";
  } else {  
    $sth = $ip_dbh->prepare("select pool from poolips where ip='$webvar{ip}'");
    $sth->execute;
    my @data = $sth->fetchrow_array;
    print "$webvar{ip} in $data[0] updated\n<hr>\n";
    syslog "notice", "$authuser updated pool IP $webvar{ip}";
  }
} elsif ($webvar{action} eq 'showusers') {
  print "Notes:<br>\n".
	"<li>Admin users automatically get all other priviledges.\n".
	"<li>Everyone has basic read access.\n".
	"<hr>Add new user:<form action=admin.cgi method=POST>\n".
	"Username: <input name=username><br>\n".
	"Password: <input name=password> <input type=checkbox name=preenc>Password is pre-encrypted (MUST be crypt() encrypted)<br>\n".
	"<input type=submit value='Add user'><input type=hidden name=action value=newuser></form>\n";

  print "<hr>Users with access:\n<table border=1>\n";
  print "<tr><td></td><td align=center colspan=3>General access</td></tr>\n";
  print "<tr><td>Username</td><td>Add new</td><td>Change</td>".
	"<td>Delete</td><td>Systems/Networking</td><td>Admin user</td></tr>\n".
	"<form action=admin.cgi method=POST>\n";
  $sth = $ip_dbh->prepare("select username,acl from users order by username");
  $sth->execute;
  while (my @data = $sth->fetchrow_array) {
    print "<form action=admin.cgi method=POST><input type=hidden name=action value=updacl>".
	qq(<tr><td>$data[0]<input type=hidden name=username value="$data[0]"></td><td>).
    # Now for the fun bit.  We have to pull apart the ACL field and
    # output a bunch of checkboxes.
    	"<input type=checkbox name=add".($data[1] =~ /a/ ? ' checked=y' : '').
	"></td><td><input type=checkbox name=change".($data[1] =~ /c/ ? ' checked=y' : '').
	"></td><td><input type=checkbox name=del".($data[1] =~ /d/ ? ' checked=y' : '').
	"></td><td><input type=checkbox name=sysnet".($data[1] =~ /s/ ? ' checked=y' : '').
	"></td><td><input type=checkbox name=admin".($data[1] =~ /A/ ? ' checked=y' : '').
	qq(></td><td><input type=submit value="Update"></td></form>\n).
	"<form action=admin.cgi method=POST><td><input type=hidden name=action value=deluser>".
	"<input type=hidden name=username value=$data[0]>".
	qq(<input type=submit value="Delete user"></tr></form>\n);

  }
  print "</table>\n";
} elsif ($webvar{action} eq 'updacl') {
  print "Updating ACL for $webvar{username}:<br>\n";
  my $acl = 'b';
  if ($webvar{admin} eq 'on') {
    $acl .= "acdsA";
  } else {
    $acl .= ($webvar{add} eq 'on' ? 'a' : '').
	($webvar{change} eq 'on' ? 'c' : '').
	($webvar{del} eq 'on' ? 'd' : '').
	($webvar{sysnet} eq 'on' ? 's' : '');
  }
  print "New ACL: $acl<br>\n";

  $sth = $ip_dbh->prepare("update users set acl='$acl' where username='$webvar{username}'");
  $sth->execute;
  print "OK\n" if !$sth->err;

  print qq(<hr><a href="admin.cgi?action=showusers">Back</a> to user listing\n);

} elsif ($webvar{action} eq 'newuser') {
  print "Adding user $webvar{username}...\n";
  my $cr_pass = ($webvar{preenc} ? $webvar{password} :
	crypt $webvar{password}, join('',('.','/',0..9,'A'..'Z','a'..'z')[rand 64, rand 64]));
  $sth = $ip_dbh->prepare("insert into users (username,password,acl) values ".
	"('$webvar{username}','$cr_pass','b')");
  $sth->execute;
  if ($sth->err) {
    print "<br>Error adding user: ".$sth->errstr;
  } else {
    print "OK\n";
  }

  print qq(<hr><a href="admin.cgi?action=showusers">Back</a> to user listing\n);

} elsif ($webvar{action} eq 'deluser') {
  print "Deleting user $webvar{username}.<br>\n";
  $sth = $ip_dbh->prepare("delete from users where username='$webvar{username}'");
  $sth->execute;
  print "OK\n" if !$sth->err;

  print qq(<hr><a href="admin.cgi?action=showusers">Back</a> to user listing\n);

} elsif ($webvar{action} eq 'emailnotice') {
  print "<h4>Email notice management:</h4>\nClick the email addresses to edit that list.";
  $sth = $ip_dbh->prepare("SELECT action,reciplist FROM notify");
  $sth->execute;

  print "<table border=1>\n";
  while (my ($notice_code,$reciplist) = $sth->fetchrow_array() ) {
##fixme: hairy mess, only a few things call mailNotify() anyway, so many possible notices won't work.
    my $action_out = dispNoticeCode($notice_code);
    print "<tr><td>$action_out</td>".
	qq(<td><a href="admin.cgi?action=ednotice&code=$notice_code">$reciplist</a></td>).
	qq(<td><a href="admin.cgi?action=delnotice&code=$notice_code">Delete</a></tr>\n);
  }
  print qq(<tr><td colspan=2>Known "special" codes:<br>
<ul style="margin-top: 0px; margin-bottom: 0px;">
	<li>swi: Notify if block being updated has SWIP flag set</li>
</ul></td></tr>
</table>
);

# add new entries from this tangle:
  print "<h4>Add new notification:</h4>\n".
	"Note:  Failure notices on most conditions are not yet supported.\n";

  print qq(<table border=1><form action=admin.cgi method="POST">
<input type=hidden name=action value=addnotice>
<tr>
<td>Recipients</td><td colspan=3><textarea name=reciplist cols=50 rows=5></textarea></td></tr>
<tr><td>Action</td><td>
	<table><tr>
		<td><input type=radio name=msgaction value=a>Add &nbsp;
		<input type=radio name=msgaction value=u>Update &nbsp;
		<input type=radio name=msgaction value=d>Delete &nbsp;
		<input type=radio name=msgaction value=n>New listitem</td>
	</tr><tr>
		<td>
		<input type=radio name=msgaction value=s:>Special: <input name=special>(requires code changes)
	</td></tr></table>
</td>
<td>Failure?</td><td><input type=checkbox name=onfail></td></tr>
<tr><td>Event/Allocation type:</td><td colspan=3>
	<table>
	<tr>
		<td><input type=radio name=alloctype value=a>All allocations</td>
		<td><input type=radio name=alloctype value=.i>All static IPs</td>
		<td><input type=radio name=alloctype value=ci>New city</td>
		<td><input type=radio name=alloctype value=no>New node</td>
	</tr>
	<tr>
);

  $sth = $ip_dbh->prepare("SELECT type,dispname FROM alloctypes WHERE listorder < 500 ".
	"ORDER BY listorder");
  $sth->execute;
  my $i=0;
  while (my ($type,$disp) = $sth->fetchrow_array) {
    print "		<td><input type=radio name=alloctype value=$type>$disp</td>";
    $i++;
    print "	</tr>\n\t<tr>"
	if ($i % 4 == 0);
  }

  print qq(	</tr>
	</table>
</tr>
<tr><td colspan=4 align=center><input type=submit value="Add notice"></td></tr>
</table>
</form>
);
  ## done spitting out add-new-spam-me-now table

} elsif ($webvar{action} eq 'addnotice') {
  $webvar{alloctype} = $webvar{special} if $webvar{msgaction} eq 's:';
  if ($webvar{msgaction} && $webvar{alloctype} && $webvar{reciplist}) {
    $webvar{reciplist} =~ s/[\r\n]+/,/g;
    $webvar{msgaction} = "f:$webvar{msgaction}" if $webvar{onfail};
    print "Adding notice to $webvar{reciplist} for ".dispNoticeCode($webvar{msgaction}.$webvar{alloctype}).":\n";
    $sth = $ip_dbh->prepare("INSERT INTO notify (action, reciplist) VALUES (?,?)");
##fixme:  automagically merge reciplists iff action already exists
    $sth->execute($webvar{msgaction}.$webvar{alloctype}, $webvar{reciplist});
    if ($sth->err) {
      print "Failed:  DB error: ".$sth->errstr."\n";
    } else {
      print "OK!<br>\n"
    }
  } else {
    print "Need to specify at least one recipient, an action, and an allocation type. ".
	qq{("Special" content is considered an allocation type).  Hit the Back button and try again.<br>\n};
  }
  print qq(<a href="admin.cgi?action=emailnotice">Back to email notice list</a>\n);

} elsif ($webvar{action} eq 'delnotice') {
  print "Deleting notices on ".dispNoticeCode($webvar{code}.$webvar{alloctype}).":\n";
  $sth = $ip_dbh->prepare("DELETE FROM notify WHERE action=?");
  $sth->execute($webvar{code});
  if ($sth->err) {
    print "Failed:  DB error: ".$sth->errstr."\n";
  } else {
    print "OK!<br>\n"
  }
  print qq(<a href="admin.cgi?action=emailnotice">Back to email notice list</a>\n);

} elsif ($webvar{action} eq 'ednotice') {
  print "<h4>Editing recipient list for '".dispNoticeCode($webvar{code})."':</h4>\n";
  $sth = $ip_dbh->prepare("SELECT reciplist FROM notify WHERE action=?");
  $sth->execute($webvar{code});
  my ($reciplist) = $sth->fetchrow_array;
  $reciplist =~ s/,/\n/g;
  print qq(<form action=admin.cgi method=POST><input type=hidden name=code value="$webvar{code}">\n).
	qq(<input type=hidden name=action value="updnotice"><table border=1><tr><td>).
	qq(<textarea cols="40" rows="5" name=reciplist>$reciplist</textarea></td><td><input type=submit value="Update">\n).
	"</td></tr></table></form>\n";
} elsif ($webvar{action} eq 'updnotice') {
  print "<h4>Updating recipient list for '".dispNoticeCode($webvar{code})."':</h4>\n";
  $sth = $ip_dbh->prepare("UPDATE notify SET reciplist=? WHERE action=?");
  $webvar{reciplist} =~ s/[\r\n]+/,/g;
  $sth->execute($webvar{reciplist}, $webvar{code});
  if ($sth->err) {
    print "Failed:  DB error: ".$sth->errstr."\n";
  } else {
    print "OK!<br>\n"
  }
  print qq(<a href="admin.cgi?action=emailnotice">Back to email notice list</a>\n);
} elsif ($webvar{action} ne '<NULL>') {
  print "webvar{action} check failed: Don't know how to $webvar{action}";
}

print "Content-type: text/html\n\n".$header->output;
print $page->output;

##fixme:  make me a footer param!
print qq(<hr><div><a href="/ip/">Back</a> to main interface</div>\n);

# We print the footer here, so we don't have to do it elsewhere.
my $footer = HTML::Template->new(filename => "footer.tmpl");
# we're already in the admin tools, no need to provide a bottom link.  maybe.
#$footer->param(adminlink => ($IPDBacl{$authuser} =~ /A/));

print $footer->output;

$ip_dbh->disconnect;

exit;


# Hokay.  This is a little different.  We have a few specific functions here:
#  -> Assign arbitrary subnet from arbitrary free space
#  -> Tweak individual DB fields
#


# Tweak allocfrom into shape.
sub fix_allocfrom {
  if ($webvar{allocfrom} =~ /^(\d+\.){2}\d+$/) {
    # 3-octet class C specified
    $webvar{allocfrom} .= ".0/24";
  } elsif ($webvar{allocfrom} =~ /^(\d+\.){3}\d+$/) {
    # 4-octet IP specified;  
    $webvar{allocfrom} .= "/24";
  }
}


# List free blocks in a /24 for arbitrary manual allocation
sub showfree($) {
  my $cidr = new NetAddr::IP $_[0];
  print "Showing free blocks in $cidr<br>\n".
	"<table border=1>\n";
  $sth = $ip_dbh->prepare("select * from freeblocks where cidr <<= '$cidr' order by cidr");
  $sth->execute;
  while (my @data = $sth->fetchrow_array) {
    my $temp = new NetAddr::IP $data[0];
    print "<tr><form action=admin.cgi method=POST><input type=hidden name=action value=assign>\n".
	qq(<td>$temp<input type=hidden name=allocfrom value="$temp"></td>\n).
	"<td>".
	(($temp->masklen == 30) ? '<input type=hidden name=masklen value=30>30'
	  : "<select name=masklen><option>30</option>\n<option>29</option>\n") .
	(($temp->masklen < 29) ? "<option>28</option>\n" : '') .
	(($temp->masklen < 28) ? "<option>27</option>\n" : '') .
	(($temp->masklen < 27) ? "<option>26</option>\n" : '') .
	(($temp->masklen < 26) ? "<option>25</option>\n" : '') .
	(($temp->masklen < 25) ? "<option>24</option>\n" : '') .
	"</td>".
	qq(<td>$data[2]</td><td><input type=submit value="Allocate from here"></td>).
	"\n</form></tr>\n";
  }
  print "</table>\n";
}


# Show allocations to allow editing.
sub showAllocs($) {
  my $cidr = new NetAddr::IP $_[0];
  print "Edit custID, allocation type, city for allocations in ".
	"$cidr:\n<table border=1>";
  $sth = $ip_dbh->prepare("select * from allocations where cidr <<= '$cidr' order by cidr");
  $sth->execute;
  while (my @data = $sth->fetchrow_array) {
    print "<tr><form action=admin.cgi method=POST><input type=hidden name=action value=update>\n".
	qq(<td>$data[0]<input type=hidden value="$data[0]" name=block></td>\n).
	qq(<td><input name=custid value="$data[1]"></td>\n).
	"<td><select name=alloctype>";

    my $sth2 = $ip_dbh->prepare("select type,listname from alloctypes".
	" where listorder < 500 and not (type like '_i') order by listorder");
    $sth2->execute;
    while (my @types = $sth2->fetchrow_array) {
      print "<option". (($data[2] eq $types[0]) ? ' selected' : '') .
	" value='$types[0]'>$types[1]</option>\n";
    }

    print qq(<td><input name=city value="$data[3]"></td>\n).
	"<td>$data[4]</td><td>$data[5]</td>".
	qq(<td><input type=submit value="Update"></td></form></tr>\n);
  }
  print "</table>\n";

  # notes
  print "<hr><b>Notes:</b>\n".
	"<ul>\n<li>Use the main interface to update description and notes fields\n".
	"<li>Changing the allocation type here will NOT affect IP pool data.\n".
	"</ul>\n";
}


# Stuff updates into DB
sub update {
  eval {
    # Relatively simple SQL transaction here.  Note that we're deliberately NOT
    # updating notes/desc here as it's available through the main interface.
    $sth = $ip_dbh->prepare("update allocations set custid='$webvar{custid}',".
	"city='$webvar{city}',type='$webvar{alloctype}' where cidr='$webvar{block}'");
    $sth->execute;
    $ip_dbh->commit;
  };
  if ($@) {
    carp "Transaction aborted because $@";
    eval { $ip_dbh->rollback; };
    syslog "err", "$authuser could not update block '$webvar{block}': '$@'";
  } else {
    # If we get here, the operation succeeded.
    syslog "notice", "$authuser updated $webvar{block}";
    print "Allocation $webvar{block} updated<hr>\n";
  }
  # need to get /24 that block is part of
  my @bits = split /\./, $webvar{block};
  $bits[3] = "0/24";
  showAllocs((join ".", @bits));
}


# showPool()
# List all IPs in a pool, and allow arbitrary admin changes to each
# Allow changes to ALL fields
sub showPool($) {
  my $pool = new NetAddr::IP $_[0];
  print qq(Listing pool $pool:\n<table border=1>
<form action=admin.cgi method=POST>
<input type=hidden name=action value=updatepool>
<tr><td align=right>Customer ID:</td><td><input name=custid></td></tr>
<tr><td align=right>Customer location:</td><td><input name=city></td></tr>
<tr><td align=right>Type:</td><td><select name=type><option selected>-</option>\n);

  $sth = $ip_dbh->prepare("select type,listname from alloctypes where type like '_i' order by listorder");
  $sth->execute;
  while (my @data = $sth->fetchrow_array) {
    print "<option value='$data[0]'>$data[1]</option>\n";
  }

  print qq(</select></td></tr>
<tr><td align=right>Available?</td><td><input type=checkbox value=y></td></tr>
<tr><td align=right>Description/name:</td><td><input name=desc size=40></td></tr>
<tr><td align=right>Notes:</td><td><textarea name=notes rows=3 cols=40></textarea></td></tr>
<tr><td colspan=2 align=center><input type=submit value="Update"></td></tr>
).
	"</table>Update the following record:<table border=1>\n";
  $sth = $ip_dbh->prepare("select pool,ip,custid,city,type,available,description,notes from poolips where pool='$pool' order by ip");
  $sth->execute;
  while (my @data = $sth->fetchrow_array) {
    print qq(<tr><td><input type=radio name=ip value="$data[1]">$data[1]</td>).
	"<td>$data[2]</td><td>$data[3]</td><td>$data[4]</td>".
	"<td>$data[5]</td><td>$data[6]</td><td>$data[7]</td></tr>\n";
  }
  print "</form></table>\n";
}


# interpret the notify codes
sub dispNoticeCode {
  my $code = shift;
  my $action_out = '';

  if ($code =~ /^s:/) {
    $code =~ s/^s:/Special: /;
    return $code;
  }
  if ($code =~ /^f:(.+)$/) {
    $code =~ s/^f://;
    $action_out = "Failure on ";
  }
  if (my $target = $code =~ /^n(.+)/) {
    $action_out .= "New ";
    if ($1 eq 'ci') { $action_out .= "city"; }
    elsif ($1 eq 'no') { $action_out .= "node"; }
    else { $action_out .= '&lt;unknown&gt;'; }
  } else {
    my ($action,$target) = ($code =~ /^(.)(.+)$/);
    if ($action eq 'a')      { $action_out .= 'Add '; }
    elsif ($action eq 'u')   { $action_out .= 'Update '; }
    elsif ($action eq 'd')   { $action_out .= 'Delete '; }
##fixme:  what if we get something funky?
# What about the eleventy-billion odd combinations possible?
# this should give an idea of the structure tho
    if ($target eq 'a') { $action_out .= "all"; }
    elsif ($target eq '.i') {
      $action_out .= "all static IPs";
    }
    else { $action_out .= $disp_alloctypes{$target}; }
  }
  return $action_out;
}
